@@ -6,68 +6,69 @@ import os
 import subprocess
 import shlex
 import sys
+from configparser import RawConfigParser
 from decimal import Decimal
+from functools import lru_cache
 from pathlib import Path
 from platform import platform
 from tempfile import NamedTemporaryFile
 
 try:
-    from ConfigParser import RawConfigParser
-except ImportError:
-    from configparser import RawConfigParser
-
-try:
-    import thread
     import threading
 except ImportError:
-    thread = None
+    threading = None
 
 import argcomplete
 import colorama
 from appdirs import user_config_dir
+from halo import Halo
+from spinners.spinners import Spinners
 
-from .ansi import ANSIEmojiLoglevelFormatter, ANSIStrippingEmojiLoglevelFormatter, ANSIStrippingFormatter, ansi_colors, ansi_escape, format_ansi
-from .configuration import Configuration, SubparserWrapper, get_argument_name, handle_store_boolean
+from .ansi import MILCFormatter, ansi_colors, ansi_config, ansi_escape, format_ansi
 from .attrdict import AttrDict
+from .configuration import Configuration, SubparserWrapper, get_argument_name, get_argument_strings, handle_store_boolean
 
 
 class MILC(object):
     """MILC - An Opinionated Batteries Included Framework
     """
-    def __init__(self, version='unknown'):
+    def __init__(self, name, version, author):
         """Initialize the MILC object.
-
-            version
-                The version string to associate with your CLI program
         """
         # Setup a lock for thread safety
-        self._lock = threading.RLock() if thread else None
+        self._lock = threading.RLock() if threading else None
 
         # Define some basic info
         self.acquire_lock()
+        self.prog_name = name
+        self.version = version
+        self.author = author
         self._config_store_true = []
         self._config_store_false = []
         self._description = None
         self._entrypoint = None
+        self._spinners = {}
+        self._subcommand = None
         self._inside_context_manager = False
         self.ansi = ansi_colors
         self.arg_only = {}
-        self.config = self.config_source = None
-        self.config_file = None
+        self.config_file = self.find_config_file()
         self.default_arguments = {}
-        self.version = version
         self.platform = platform()
-
-        # Figure out our program name
-        self.prog_name = sys.argv[0][:-3] if sys.argv[0].endswith('.py') else sys.argv[0]
-        self.prog_name = os.environ.get('MILC_APP_NAME', os.path.basename(self.prog_name))
+        self.interactive = sys.stdin.isatty()
         self.release_lock()
+        self._deprecated_arguments = {}
+        self._deprecated_commands = {}
 
         # Initialize all the things
         self.initialize_config()
         self.initialize_argparse()
         self.initialize_logging()
 
+    @property
+    def config_dir(self):
+        return self.config_file.parent
+
     @property
     def description(self):
         return self._description
@@ -79,7 +80,7 @@ class MILC(object):
     def echo(self, text, *args, **kwargs):
         """Print colorized text to stdout.
 
-        ANSI color strings (such as {fg-blue}) will be converted into ANSI
+        ANSI color strings (such as {fg_blue}) will be converted into ANSI
         escape sequences, and the ANSI reset sequence will be added to all
         strings.
 
@@ -123,12 +124,18 @@ class MILC(object):
         if not capture_output and combined_output:
             raise ValueError("Can't use capture_output=False and combined_output=True at the same time.")
 
-        # On some windows platforms (msys2, possibly others) you have to execute the command through a subshell.
+        # On some windows platforms (msys2, possibly others) you have to
+        # execute the command through a subshell. As well, after execution
+        # stdin is broken so things like milc.questions no longer work.
+        # We pass `stdin=subprocess.DEVNULL` by default to prevent that.
         if 'windows' in self.platform.lower():
             safecmd = map(shlex.quote, command)
             safecmd = ' '.join(safecmd)
             command = [os.environ['SHELL'], '-c', safecmd]
 
+            if 'stdin' not in kwargs:
+                kwargs['stdin'] = subprocess.DEVNULL
+
         # Argument Processing
         if capture_output:
             kwargs['stdout'] = subprocess.PIPE
@@ -137,6 +144,9 @@ class MILC(object):
         if combined_output:
             kwargs['stderr'] = subprocess.STDOUT
 
+        if 'stdin' in kwargs and kwargs['stdin'] is None:
+            del kwargs['stdin']
+
         if text:
             kwargs['universal_newlines'] = True
 
@@ -158,6 +168,7 @@ class MILC(object):
         self._subparsers = None
         self.argwarn = argcomplete.warn
         self.args = AttrDict()
+        self.args_passed = AttrDict()
         self._arg_parser = argparse.ArgumentParser(**kwargs)
         self.set_defaults = self._arg_parser.set_defaults
         self.release_lock()
@@ -165,23 +176,23 @@ class MILC(object):
     def print_help(self, *args, **kwargs):
         """Print a help message for the main program or subcommand, depending on context.
         """
-        if self._entrypoint.__name__ in self.subcommands:
-            return self.subcommands[self._entrypoint.__name__].print_help(*args, **kwargs)
+        if self._subcommand:
+            return self.subcommands[self._subcommand.__name__].print_help(*args, **kwargs)
 
         return self._arg_parser.print_help(*args, **kwargs)
 
     def print_usage(self, *args, **kwargs):
         """Print brief description of how the main program or subcommand is invoked, depending on context.
         """
-        if self._entrypoint.__name__ in self.subcommands:
-            return self.subcommands[self._entrypoint.__name__].print_usage(*args, **kwargs)
+        if self._subcommand:
+            return self.subcommands[self._subcommand.__name__].print_usage(*args, **kwargs)
 
         return self._arg_parser.print_usage(*args, **kwargs)
 
-    def completer(self, completer):
-        """Add an argcomplete completer to this subcommand.
+    def log_deprecated_warning(self, item_type, name, reason):
+        """Logs a warning with a custom message if a argument or command is deprecated.
         """
-        self._arg_parser.completer = completer
+        self.log.warning("Warning: %s '%s' is deprecated:\n\t%s", item_type, name, reason)
 
     def add_argument(self, *args, **kwargs):
         """Wrapper to add arguments and track whether they were passed on the command line.
@@ -189,12 +200,37 @@ class MILC(object):
         if 'action' in kwargs and kwargs['action'] == 'store_boolean':
             return handle_store_boolean(self, *args, **kwargs)
 
+        arg_name = get_argument_name(self._arg_parser, *args, **kwargs)
+        arg_strings = get_argument_strings(self._arg_parser, *args, **kwargs)
+
+        # Handle tab completion
+        completer = kwargs.get('completer')
+
+        if 'completer' in kwargs:
+            del kwargs['completer']
+
         self.acquire_lock()
 
-        self._arg_parser.add_argument(*args, **kwargs)
+        if completer:
+            self._arg_parser.add_argument(*args, **kwargs).completer = completer
+        else:
+            self._arg_parser.add_argument(*args, **kwargs)
+
+        # Record the default for this argument
         if 'general' not in self.default_arguments:
             self.default_arguments['general'] = {}
-        self.default_arguments['general'][get_argument_name(self, *args, **kwargs)] = kwargs.get('default')
+
+        self.default_arguments['general'][arg_name] = kwargs.get('default')
+
+        # Determine if it was passed on the command line
+        if 'general' not in self.args_passed:
+            self.args_passed['general'] = {}
+
+        self.args_passed['general'][arg_name] = False
+
+        for arg in arg_strings:
+            if arg in sys.argv:
+                self.args_passed['general'][arg_name] = True
 
         self.release_lock()
 
@@ -222,7 +258,9 @@ class MILC(object):
         self.add_argument('--log-file-fmt', default='[%(levelname)s] [%(asctime)s] [file:%(pathname)s] [line:%(lineno)d] %(message)s', help='Format string for log file.')
         self.add_argument('--log-file-level', default='info', choices=['debug', 'info', 'warning', 'error', 'critical'], help='Logging level for log file.')
         self.add_argument('--log-file', help='File to write log messages to')
-        self.add_argument('--color', action='store_boolean', default=True, help='color in output')
+        self.add_argument('--color', action='store_boolean', default=ansi_config['color'], help='color in output')
+        self.add_argument('--unicode', action='store_boolean', default=ansi_config['unicode'], help='unicode loglevels')
+        self.add_argument('--interactive', action='store_true', help='Force interactive mode even when stdout is not a tty.')
         self.add_argument('--config-file', help='The location for the configuration file')
         self.arg_only['config_file'] = ['general']
 
@@ -246,13 +284,14 @@ class MILC(object):
         if self._lock:
             self._lock.release()
 
+    @lru_cache(maxsize=None)
     def find_config_file(self):
         """Locate the config file.
         """
         if '--config-file' in sys.argv:
             return Path(sys.argv[sys.argv.index('--config-file') + 1]).expanduser().resolve()
 
-        filedir = user_config_dir(appname=self.prog_name, appauthor=os.environ.get('MILC_APP_AUTHOR', self.prog_name.upper()))
+        filedir = user_config_dir(appname=self.prog_name, appauthor=self.author)
         filename = '%s.ini' % self.prog_name
 
         return Path(filedir, filename).resolve()
@@ -264,16 +303,42 @@ class MILC(object):
             raise RuntimeError('You must run this before the with statement!')
 
         def argument_function(handler):
-            subcommand_name = handler.__name__.replace("_", "-")
+            config_name = handler.__name__
+            subcommand_name = config_name.replace("_", "-")
+            arg_name = get_argument_name(self._arg_parser, *args, **kwargs)
+            arg_strings = get_argument_strings(self._arg_parser, *args, **kwargs)
+
+            if 'deprecated' in kwargs:
+                self._deprecated_arguments[arg_name] = kwargs['deprecated']
+                if kwargs['help']:
+                    kwargs['help'] += f" [Deprecated]: {kwargs['deprecated']}"
+                else:
+                    kwargs['help'] = f"[Deprecated]: {kwargs['deprecated']}"
+                del kwargs['deprecated']
 
             if kwargs.get('arg_only'):
-                arg_name = get_argument_name(self, *args, **kwargs)
-
                 if arg_name not in self.arg_only:
                     self.arg_only[arg_name] = []
 
                 self.arg_only[arg_name].append(handler.__name__)
                 del kwargs['arg_only']
+            else:
+                if arg_name not in self.default_arguments:
+                    self.default_arguments[config_name] = {}
+
+                self.default_arguments[config_name][arg_name] = kwargs.get('default')
+
+                if self.config[config_name][arg_name] is None:
+                    self.config[config_name][arg_name] = kwargs.get('default')
+
+                if config_name not in self.args_passed:
+                    self.args_passed[config_name] = {}
+
+                self.args_passed[config_name][arg_name] = False
+
+                for arg in arg_strings:
+                    if arg in sys.argv:
+                        self.args_passed[config_name][arg_name] = True
 
             if handler is self._entrypoint:
                 self.add_argument(*args, **kwargs)
@@ -302,19 +367,19 @@ class MILC(object):
             self.args[key] = value
 
         if 'entrypoint' in self.args:
-            self._entrypoint = self.args.entrypoint
+            self._subcommand = self.args.entrypoint
 
         self.release_lock()
 
-    def read_config_file(self, config_file):
+    def read_config_file(self):
         """Read in the configuration file and return Configuration objects for it and the config_source.
         """
         config = Configuration()
         config_source = Configuration()
 
-        if config_file.exists():
+        if self.config_file.exists():
             raw_config = RawConfigParser()
-            raw_config.read(str(config_file))
+            raw_config.read(str(self.config_file))
 
             # Iterate over the config file options and write them into config
             for section in raw_config.sections():
@@ -343,53 +408,31 @@ class MILC(object):
         """Read in the configuration file and store it in self.config.
         """
         self.acquire_lock()
-        self.config_file = self.find_config_file()
-        self.config, self.config_source = self.read_config_file(self.config_file)
+        self.config, self.config_source = self.read_config_file()
         self.release_lock()
 
     def merge_args_into_config(self):
         """Merge CLI arguments into self.config to create the runtime configuration.
         """
         self.acquire_lock()
+        subcommand_name = self._subcommand.__name__ if self._subcommand else None
+
         for argument in self.args:
             if argument in ('subparsers', 'entrypoint'):
                 continue
 
             # Find the argument's section
-            # Underscores in command's names are converted to dashes during initialization.
-            # TODO(Erovia) Find a better solution
-            entrypoint_name = self._entrypoint.__name__.replace("_", "-")
-            if entrypoint_name in self.default_arguments and argument in self.default_arguments[entrypoint_name]:
-                argument_found = True
-                section = self._entrypoint.__name__
             if argument in self.default_arguments['general']:
-                argument_found = True
                 section = 'general'
-
-            if not argument_found:
-                raise RuntimeError('Could not find argument in `self.default_arguments`. This should be impossible!')
-                exit(1)
+            else:
+                section = subcommand_name
 
             if argument not in self.arg_only or section not in self.arg_only[argument]:
                 # Determine the arg value and source
                 arg_value = getattr(self.args, argument)
-                passed_on_cmdline = False
-
-                if section in self.subcommands:
-                    default_value = self.subcommands[section].get_default(argument)
-                else:
-                    default_value = self._arg_parser.get_default(argument)
-
-                if argument in self._config_store_true and arg_value:
-                    passed_on_cmdline = True
-                elif argument in self._config_store_false and not arg_value:
-                    passed_on_cmdline = True
-                elif arg_value is not None:
-                    if self.config[section][argument] is None or arg_value != default_value:
-                        passed_on_cmdline = True
 
                 # Merge this argument into self.config
-                if passed_on_cmdline and (argument in self.default_arguments['general'] or argument in self.default_arguments[entrypoint_name] or argument not in self.config[entrypoint_name]):
+                if (self.args_passed[section][argument] or (self.config[section][argument] is None) or (argument in self._config_store_true and arg_value) or (argument in self._config_store_false and not arg_value)):
                     self.config[section][argument] = arg_value
                     self.config_source[section][argument] = 'argument'
 
@@ -403,19 +446,15 @@ class MILC(object):
         for section_name, section in config.items():
             sane_config.add_section(section_name)
             for option_name, value in section.items():
-                if section_name == 'general':
-                    if option_name in ['config_file']:
-                        continue
-                if value is not None:
+                if self.config_source[section_name][option_name] == 'config_file' and value is not None:
                     sane_config.set(section_name, option_name, str(value))
 
-        config_dir = self.config_file.parent
-        if not config_dir.exists():
-            config_dir.mkdir(parents=True, exist_ok=True)
+        if not self.config_dir.exists():
+            self.config_dir.mkdir(parents=True, exist_ok=True)
 
         # Write the config file atomically.
         self.acquire_lock()
-        with NamedTemporaryFile(mode='w', dir=str(config_dir), delete=False) as tmpfile:
+        with NamedTemporaryFile(mode='w', dir=str(self.config_dir), delete=False) as tmpfile:
             sane_config.write(tmpfile)
 
         if os.path.getsize(tmpfile.name) > 0:
@@ -431,7 +470,7 @@ class MILC(object):
             self.log.warning('%s.config_file not set, not saving config!', self.__class__.__name__)
             return
 
-        config, config_source = self.read_config_file(self.config_file)
+        config, config_source = self.read_config_file()
 
         if section in config and option in config[section] and config[section][option] is None:
             del config[section][option]
@@ -456,6 +495,25 @@ class MILC(object):
         self._save_config_file(self.config)
         self.log.info('Wrote configuration to %s', shlex.quote(str(self.config_file)))
 
+    def check_deprecated(self):
+        entry_name = self._entrypoint.__name__
+
+        if entry_name in self._deprecated_commands:
+            msg = self._deprecated_commands[entry_name]
+            self.log_deprecated_warning('Entrypoint', entry_name, msg)
+
+        if self._subcommand:
+            name = self._subcommand.__name__.replace("_", "-")
+            if name in self._deprecated_commands:
+                msg = self._deprecated_commands[name]
+                self.log_deprecated_warning('Subcommand', name, msg)
+
+        deprecated_args_passed = [arg.replace('-', '') for arg in sys.argv if arg.replace('-', '') in self._deprecated_arguments]
+
+        for arg in deprecated_args_passed:
+            msg = self._deprecated_arguments[arg]
+            self.log_deprecated_warning('Argument', arg, msg)
+
     def __call__(self):
         """Execute the entrypoint function.
         """
@@ -464,13 +522,23 @@ class MILC(object):
             with self:
                 return self.__call__()
 
-        if not self._entrypoint:
-            raise RuntimeError('No entrypoint provided!')
+        self.check_deprecated()
 
-        return self._entrypoint(self)
+        if self._subcommand:
+            return self._subcommand(self)
+        elif self._entrypoint:
+            return self._entrypoint(self)
 
-    def entrypoint(self, description):
-        """Decorator that marks the entrypoint for simple scripts without subcommands.
+        raise RuntimeError('No entrypoint provided!')
+
+    def entrypoint(self, description, deprecated=None):
+        """Decorator that marks the entrypoint used when a subcommand is not supplied.
+        Args:
+            description
+                A one-line description to display in --help
+
+            deprecated
+                Deprecation message. When set the subcommand will marked as deprecated and this message will be displayed in the help output.
         """
         if self._inside_context_manager:
             raise RuntimeError('You must run this before cli()!')
@@ -481,6 +549,12 @@ class MILC(object):
 
         def entrypoint_func(handler):
             self.acquire_lock()
+
+            if deprecated:
+                name = handler.__name__
+                self._deprecated_commands[name] = deprecated
+                self.description += f' [Deprecated]: {deprecated}'
+
             self._entrypoint = handler
             self.release_lock()
 
@@ -488,7 +562,7 @@ class MILC(object):
 
         return entrypoint_func
 
-    def add_subcommand(self, handler, description, hidden=False, **kwargs):
+    def add_subcommand(self, handler, description, hidden=False, deprecated=None, **kwargs):
         """Register a subcommand.
 
         Args:
@@ -501,6 +575,10 @@ class MILC(object):
 
             hidden
                 When True don't display this command in --help
+
+            deprecated
+                Deprecation message. When set the subcommand will be marked as deprecated
+                and this message will be displayed in help output.
         """
         if self._inside_context_manager:
             raise RuntimeError('You must run this before the with statement!')
@@ -510,6 +588,10 @@ class MILC(object):
 
         name = handler.__name__.replace("_", "-")
 
+        if deprecated:
+            self._deprecated_commands[name] = deprecated
+            description += f' [Deprecated]: {deprecated}'
+
         self.acquire_lock()
         if not hidden:
             self._subparsers.metavar = "{%s,%s}" % (self._subparsers.metavar[1:-1], name) if self._subparsers.metavar else "{%s%s}" % (self._subparsers.metavar[1:-1], name)
@@ -549,14 +631,13 @@ class MILC(object):
         if self.config.general.verbose:
             self.log_print_level = logging.DEBUG
 
-        self.log_file = self.config.general.log_file or self.log_file
-        self.log_file_format = ANSIStrippingFormatter(self.config.general.log_file_fmt, self.config.general.datetime_fmt)
-        self.log_file_level = getattr(logging, self.config.general.log_file_level.upper())
+        ansi_config['color'] = self.config.general.color
+        ansi_config['unicode'] = self.config.general.unicode
 
-        if self.config.general.color:
-            self.log_format = ANSIEmojiLoglevelFormatter(self.config.general.log_fmt, self.config.general.datetime_fmt)
-        else:
-            self.log_format = ANSIStrippingEmojiLoglevelFormatter(self.config.general.log_fmt, self.config.general.datetime_fmt)
+        self.log_file = self.config.general.log_file or self.log_file
+        self.log_file_format = MILCFormatter(self.config.general.log_file_fmt, self.config.general.datetime_fmt)
+        self.log_file_level = getattr(logging, self.config.general.log_file_level.upper())
+        self.log_format = MILCFormatter(self.config.general.log_fmt, self.config.general.datetime_fmt)
 
         if self.log_file:
             self.log_file_handler = logging.FileHandler(self.log_file, self.log_file_mode)
@@ -584,6 +665,10 @@ class MILC(object):
         colorama.init()
         self.parse_args()
         self.merge_args_into_config()
+
+        if self.config.general.interactive:
+            self.interactive = True
+
         self.setup_logging()
 
         return self
@@ -597,3 +682,115 @@ class MILC(object):
             print(exc_type)
             logging.exception(exc_val)
             exit(255)
+
+    def is_spinner(self, name):
+        """Returns true if name is a valid spinner.
+        """
+        return name in Spinners.__members__ or name in self._spinners
+
+    def add_spinner(self, name, spinner):
+        """Adds a new spinner to the list of spinners.
+
+        A spinner is a dictionary with two keys:
+
+            interval
+                An integer that sets how long (in ms) to wait between frames.
+
+            frames
+                A list of frames for this spinner
+        """
+        if self.is_spinner(name):
+            raise ValueError(f'Spinner "{name}" already exists!')
+
+        if not isinstance(spinner, dict):
+            raise ValueError('The spinner must be a dictionary!')
+
+        if 'interval' not in spinner or 'frames' not in spinner:
+            raise ValueError('The spinner must have `interval` and `frames` keys!')
+
+        self._spinners[name] = spinner
+
+    def spinner(self, text, *args, spinner=None, animation='ellipsed', placement='left', color='blue', interval=-1, stream=sys.stdout, enabled=True, **kwargs):
+        """Create a spinner object for showing activity to the user.
+
+        This uses halo <https://github.com/ManrajGrover/halo> behind the scenes, most of the arguments map to Halo objects 1:1.
+
+        There are 3 basic ways to use this:
+
+        * Instantiating a spinner and then using `.start()` and `.stop()` on your object.
+        * Using a context manager (`with cli.spinner(...):`)
+        * Decorate a function (`@cli.spinner(...)`)
+
+        #### Instantiating a spinner
+
+        ```python
+        spinner = cli.spinner(text='Loading', spinner='dots')
+        spinner.start()
+
+        # Do something here
+
+        spinner.stop()
+        ```
+
+        #### Using a context manager
+
+        ```python
+        with cli.spinner(text='Loading', spinner='dots'):
+            # Do something here
+        ```
+
+        #### Decorate a function
+
+        ```python
+        @cli.spinner(text='Loading', spinner='dots')
+        def long_running_function():
+            # Do something here
+        ```
+
+        ### Arguments
+
+            text
+                The text to display next to the spinner. ANSI color strings
+                (such as {fg_blue}) will be converted into ANSI escape
+                sequences, and the ANSI reset sequence will be added to the
+                end of the string.
+
+                If *args or **kwargs are passed they will be used to
+                %-format the text.
+
+            spinner
+                The name of the spinner to use. Available names are here:
+                <https://raw.githubusercontent.com/sindresorhus/cli-spinners/dac4fc6571059bb9e9bc204711e9dfe8f72e5c6f/spinners.json>
+
+            animation
+                The animation to apply to the text if it doesn't fit the
+                terminal. One of `ellipsed`, `bounce`, `marquee`.
+
+            placement
+                Which side of the text to display the spinner on. One of
+                `left`, `right`.
+
+            color
+                Color of the spinner. One of `blue`, `grey`, `red`, `green`,
+                `yellow`, `magenta`, `cyan`, `white`
+
+            interval
+                How long in ms to wait between frames. Defaults to the spinner interval (recommended.)
+
+            stream
+                Stream to write the output. Defaults to sys.stdout.
+
+            enabled
+                Enable or disable the spinner. Defaults to `True`.
+        """
+        if isinstance(spinner, str) and spinner in self._spinners:
+            spinner = self._spinners[spinner]
+
+        return Halo(
+            text=format_ansi(text % (args or kwargs)),
+            spinner=spinner if spinner else 'line',  # FIXME: Grab one of the ascii spinners at random instead of line
+            animation=None if animation == 'ellipsed' else animation,
+            placement=placement,
+            color=color,
+            interval=interval,
+        )
